from __future__ import annotations
from typing import Dict, Optional, Union, List
from fastapi import HTTPException, Query
from aiida_gui.app.node_table import make_node_router
from aiida import orm
import traceback

project = ["id", "uuid", "time", "label", "description"]


def projected_data_to_dict(qb, project):
    """
    Convert the projected data from a QueryBuilder to a list of dictionaries.
    """
    from aiida_gui.app.utils import time_ago

    # Iterate over the results and convert each row to a dictionary
    results = []
    for row in qb.all():
        item = dict(zip(project or [], row))
        # Add computed/presentational fields
        item["pk"] = item.pop("id")
        item["ctime"] = time_ago(item.get("ctime"))
        results.append(item)
    return results


# due to a bug in aiida-core: https://github.com/aiidateam/aiida-core/pull/6828
# we need use `time` instead of `ctime`
def projected_data_to_dict_group(qb, project):
    """
    Convert the projected data from a QueryBuilder to a list of dictionaries.
    """
    from aiida_gui.app.utils import time_ago

    # Iterate over the results and convert each row to a dictionary
    results = []
    for row in qb.all():
        item = dict(zip(project or [], row))
        # Add computed/presentational fields
        item["pk"] = item.pop("id")
        item["ctime"] = time_ago(item.get("time"))
        results.append(item)
    return results


router = make_node_router(
    node_cls=orm.Group,
    prefix="groupnode",
    project=project,
    get_data_func=projected_data_to_dict_group,
    inclue_delete_route=False,
)


@router.get("/api/groupnode/{id}")
async def read_group_summary(id: int) -> Dict[str, Union[str, int]]:
    try:
        g = orm.load_group(id)
        summary = {
            "pk": g.pk,
            "uuid": g.uuid,
            "label": g.label,
            "description": g.description or "",
            "type_string": g.type_string,
            "count": g.count(),
        }
        return summary
    except Exception:
        raise HTTPException(status_code=404, detail=f"Group {id} not found")


# ---------------------------------------------------------------------------
# ─ 3. paginated members
#      GET /api/groupnode/{id}/members-data   (same contract as -data)
# ---------------------------------------------------------------------------
@router.get("/api/groupnode/{id}/members-data")
async def read_group_members(
    id: int,
    skip: int = Query(0, ge=0),
    limit: int = Query(15, gt=0, le=500),
    sortField: str = Query("pk"),
    sortOrder: str = Query("desc", pattern="^(asc|desc)$"),
    filterModel: Optional[str] = Query(None),
):

    project = ["id", "ctime", "node_type", "label", "description"]

    qb = orm.QueryBuilder()
    qb.append(
        orm.Group,
        filters={
            "id": id,
        },
        tag="group",
    )
    qb.append(
        orm.Node,
        project=project,
        with_group="group",
        tag="node",
    )

    # server‑side filters coming from the DataGrid
    if filterModel:
        from aiida_gui.app.utils import (
            translate_datagrid_filter_json,
        )

        qb.add_filter(
            "node", translate_datagrid_filter_json(filterModel, project=project)
        )

    qb.order_by({"node": {sortField: sortOrder}})
    total = qb.count()
    qb.offset(skip).limit(limit)

    results = projected_data_to_dict(qb, project)
    return {"total": total, "data": results}


@router.delete("/api/groupnode/delete" + "/{id}")
async def delete(
    id: int, dry_run: bool = False, delete_nodes: bool = False
) -> Dict[str, Union[bool, str, List[int]]]:
    from aiida.tools import delete_group_nodes

    try:
        if dry_run:
            return {
                "deleted_nodes": [id],
            }
        if delete_nodes:
            _, nodes_deleted = delete_group_nodes([id], dry_run=dry_run)
            if nodes_deleted:
                orm.Group.collection.delete(id)
                ok = True
            else:
                ok = False
        else:
            orm.Group.collection.delete(id)
            ok = True
        return {
            "deleted": ok,
            "message": (
                f"{'Deleted' if ok else 'Did not delete'} {orm.Group.__name__} {id}"
            ),
        }
    except Exception as e:
        error_traceback = traceback.format_exc()  # Capture the full traceback
        print(error_traceback)
        raise HTTPException(status_code=500, detail=str(e))


@router.delete("/api/groupnode/{group_id}/members/remove/{node_id}")
async def delete_node(
    group_id: int,
    node_id: int,
) -> Dict[str, Union[bool, str, List[int]]]:

    try:
        group = orm.load_group(group_id)
        group.remove_nodes([orm.load_node(node_id)])
        return {
            "removed": True,
            "message": f"Removed node {node_id} from the group",
        }
    except Exception as e:
        error_traceback = traceback.format_exc()  # Capture the full traceback
        print(error_traceback)
        raise HTTPException(status_code=500, detail=str(e))



import sys
import traceback
from fastapi.staticfiles import StaticFiles
from starlette.routing import Mount, Route


def get_plugins():
    import importlib.metadata

    plugins = {}

    try:
        eps = importlib.metadata.entry_points()
        if sys.version_info >= (3, 10):
            eps = eps.select(group="aiida_gui.plugins")
        else:
            eps = eps.get("aiida_gui.plugins", [])
    except Exception:
        print("Failed to get entry points")
        return plugins

    for entry in eps:
        plugin_name = entry.name
        try:
            plugin_module = entry.load()
            plugins[plugin_name] = plugin_module
        except Exception as e:
            print(traceback.format_exc())
            print(f"Failed to load plugin {plugin_name}: {e}")
            continue

    return plugins


def mount_plugins(app):
    plugins = get_plugins()
    for plugin_name, plugin_module in plugins.items():
        sub_apps = plugin_module.get("sub_apps", {})
        routers = plugin_module.get("routers", {})
        static_dirs = plugin_module.get("static_dirs", {})

        for key, sub_app in sub_apps.items():
            app.mount(
                f"/plugins/{plugin_name}/{key}", sub_app, name=f"plugin_{plugin_name}"
            )

        for key, router in routers.items():
            app.include_router(router, prefix=f"/plugins/{key}")

        for key, static_dir in static_dirs.items():
            app.mount(
                f"/plugins/{key}/static",
                StaticFiles(directory=static_dir, html=True),
                name=f"plugin_{key}",
            )


def list_routes_and_statics(app, prefix: str = ""):
    for route in app.routes:
        if isinstance(route, Mount) and isinstance(route.app, StaticFiles):
            print(
                f"[static] {prefix}{route.path}  → serves from {route.app.directory!r}"
            )
        elif isinstance(route, Mount):
            sub_prefix = prefix + route.path
            try:
                # recurse into the sub-app’s routes
                list_routes_and_statics(route.app, prefix=sub_prefix)
            except Exception:
                pass
        elif isinstance(route, Route):
            methods = ",".join(route.methods or [])
            print(f"[route ] {prefix}{route.path}  [{methods}]")



# -*- coding: utf-8 -*-
"""Declaration of FastAPI router for daemon endpoints."""
from __future__ import annotations

import typing as t

from aiida.cmdline.utils.decorators import with_dbenv
from aiida.engine.daemon.client import DaemonException, get_daemon_client
from fastapi import APIRouter, HTTPException
from pydantic import BaseModel, Field


router = APIRouter()


class DaemonStatusModel(BaseModel):
    """Response model for daemon status."""

    running: bool = Field(description="Whether the daemon is running or not.")
    num_workers: t.Optional[int] = Field(
        description="The number of workers if the daemon is running."
    )


@router.get("/api/daemon/status", response_model=DaemonStatusModel)
@with_dbenv()
async def get_daemon_status() -> DaemonStatusModel:
    """Return the daemon status."""
    client = get_daemon_client()

    if not client.is_daemon_running:
        return DaemonStatusModel(running=False, num_workers=None)

    response = client.get_numprocesses()

    return DaemonStatusModel(running=True, num_workers=response["numprocesses"])


@router.get("/api/daemon/worker")
@with_dbenv()
async def get_daemon_worker():
    """Return the daemon status."""
    client = get_daemon_client()

    if not client.is_daemon_running:
        return {}

    response = client.get_worker_info()

    return response["info"]


@router.post("/api/daemon/start", response_model=DaemonStatusModel)
@with_dbenv()
async def get_daemon_start() -> DaemonStatusModel:
    """Start the daemon."""
    client = get_daemon_client()

    if client.is_daemon_running:
        raise HTTPException(status_code=400, detail="The daemon is already running.")

    try:
        client.start_daemon()
    except DaemonException as exception:
        raise HTTPException(status_code=500, detail=str(exception)) from exception

    response = client.get_numprocesses()

    return DaemonStatusModel(running=True, num_workers=response["numprocesses"])


@router.post("/api/daemon/stop", response_model=DaemonStatusModel)
@with_dbenv()
async def get_daemon_stop() -> DaemonStatusModel:
    """Stop the daemon."""
    client = get_daemon_client()

    if not client.is_daemon_running:
        raise HTTPException(status_code=400, detail="The daemon is not running.")

    try:
        client.stop_daemon()
    except DaemonException as exception:
        raise HTTPException(status_code=500, detail=str(exception)) from exception

    return DaemonStatusModel(running=False, num_workers=None)


@router.post("/api/daemon/increase", response_model=DaemonStatusModel)
@with_dbenv()
async def increase_daemon_worker() -> DaemonStatusModel:
    """increase the daemon worker."""
    client = get_daemon_client()

    if not client.is_daemon_running:
        raise HTTPException(status_code=400, detail="The daemon is not running.")

    try:
        client.increase_workers(1)
    except DaemonException as exception:
        raise HTTPException(status_code=500, detail=str(exception)) from exception

    return DaemonStatusModel(running=False, num_workers=None)


@router.post("/api/daemon/decrease", response_model=DaemonStatusModel)
@with_dbenv()
async def decrease_daemon_worker() -> DaemonStatusModel:
    """decrease the daemon worker."""
    client = get_daemon_client()

    if not client.is_daemon_running:
        raise HTTPException(status_code=400, detail="The daemon is not running.")

    try:
        client.decrease_workers(1)
    except DaemonException as exception:
        raise HTTPException(status_code=500, detail=str(exception)) from exception

    return DaemonStatusModel(running=False, num_workers=None)



from .utils import get_node_summary_table, get_node_inputs, get_node_outputs
from aiida import orm
from fastapi import APIRouter, HTTPException
import traceback
from typing import List
from aiida.engine.processes import control

router = APIRouter()


@router.get("/api/task/{id}/{path:path}")
async def read_task(id: int, path: str):
    from aiida.orm import load_node

    try:
        node = load_node(id)
        segments = path.split("/")
        if isinstance(node, orm.WorkChainNode):
            pk = int(segments[0].split("-")[-1])
            try:
                task_node = orm.load_node(pk)
            except Exception:
                raise HTTPException(status_code=404, detail=f"Process {pk} not found")

            # metadata = get_node_summary_table(node),
            # metadata["name"] = segments[0].split("_")[0]
            # metadata["node_type"] = node.node_type
            # metadata["identifier"] = "any"
            # try:
            #     executor = inspect.getsource(task_node.process_class)
            # except Exception:
            #     executor = f"{task_node.process_class.__module__}.{task_node.process_class.__name__}"
            executor = f"{task_node.process_class.__module__}.{task_node.process_class.__name__}"

            content = {
                "node_type": task_node.node_type,
                "label": segments[0].split("_")[0],
                "metadata": get_node_summary_table(task_node),
                "inputs": get_node_inputs(task_node),
                "outputs": get_node_outputs(task_node),
                "executor": executor,
                "process": {"pk": pk},
            }
        else:
            raise HTTPException(status_code=404, detail="Node not found")
        return content
    except KeyError as e:
        error_traceback = traceback.format_exc()  # Capture the full traceback
        print(error_traceback)
        raise HTTPException(
            status_code=404, detail=f"Process {id}/{path} not found, {e}"
        )


# General function to manage task actions
async def manage_task_action(action: str, id: int, tasks: List[str]):

    node = orm.load_node(id)
    if node.is_finished:
        msg = "Process is finished. Cannot pause tasks."
        raise HTTPException(status_code=400, detail=msg)
    if isinstance(node, orm.WorkChainNode):
        print(f"Performing {action} action on tasks {tasks} in workchain {id}")
        pks = [task.split("-")[-1] for task in tasks]
        try:
            if action == "pause":
                print(f"Pausing tasks {tasks}")
                try:
                    control.pause_processes(
                        pks,
                        all_entries=None,
                        timeout=5,
                        wait=False,
                    )
                    msg = "Paused tasks"
                except Exception as e:
                    print(f"Pause processes: {pks} failed: {e}")
            elif action == "play":
                print(f"Playing tasks {tasks}")
                try:
                    control.pause_processes(
                        pks,
                        all_entries=None,
                        timeout=5,
                        wait=False,
                    )
                    msg = "Played tasks"
                except Exception as e:
                    print(f"Play processes: {pks} failed: {e}")
            elif action == "kill":
                print(f"Killing tasks {tasks}")
                try:
                    control.pause_processes(
                        pks,
                        all_entries=None,
                        timeout=5,
                        wait=False,
                    )
                    msg = "Killed tasks"
                except Exception as e:
                    print(f"Kill processes: {pks} failed: {e}")
            else:
                raise HTTPException(status_code=400, detail="Unsupported action")

            return {"message": msg}

        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))


# Endpoint for pausing tasks in a process
@router.post("/api/process/tasks/pause/{id}")
async def pause_process_tasks(id: int, tasks: List[dict] = None):
    return await manage_task_action("pause", id, tasks)


# Endpoint for playing tasks in a process
@router.post("/api/process/tasks/play/{id}")
async def play_process_tasks(id: int, tasks: List[dict] = None):
    return await manage_task_action("play", id, tasks)


# Endpoint for killing tasks in a process
@router.post("/api/process/tasks/kill/{id}")
async def kill_workgraph_tasks(id: int, tasks: List[dict] = None):
    return await manage_task_action("kill", id, tasks)



from __future__ import annotations

from typing import Dict, Optional, Union, Tuple, List, Any
from aiida.orm import load_node, Node
from datetime import datetime
from dateutil import relativedelta
from dateutil.tz import tzlocal


def get_executor_source(tdata: Any) -> Tuple[bool, Optional[str]]:
    """Get the source code of the executor."""
    import inspect
    from node_graph.executor import NodeExecutor

    executor = NodeExecutor(**tdata["executor"]).executor
    if callable(executor):
        try:
            source_lines, _ = inspect.getsourcelines(executor)
            source_code = "".join(source_lines)
            return source_code
        except (TypeError, OSError):
            source_code = tdata["executor"].get("source_code", "")
            return source_code
    else:
        return str(executor)


def get_node_recursive(links: Dict) -> Dict[str, Union[List[int], str]]:
    """Recursively get a dictionary of nodess."""
    from collections.abc import Mapping

    data = {}
    for label, value in links.items():
        if isinstance(value, Mapping):
            data.update({label: get_node_recursive(value)})
        else:
            data[label] = [value.pk, value.__class__.__name__, value.node_type]
    return data


def get_node_inputs(pk: int | Node) -> Union[str, Dict[str, Union[List[int], str]]]:
    from aiida.common.links import LinkType

    if pk is None:
        return {}

    node = load_node(pk) if isinstance(pk, int) else pk
    nodes_input = node.base.links.get_incoming(
        link_type=(LinkType.INPUT_CALC, LinkType.INPUT_WORK)
    )
    if nodes_input:
        result = get_node_recursive(nodes_input.nested())
    else:
        result = {}

    return result


def get_nodes_called(pk: int | Node) -> Union[str, Dict[str, Union[List[int], str]]]:
    from aiida.common.links import LinkType

    node = load_node(pk) if isinstance(pk, int) else pk
    links_called = node.base.links.get_outgoing(
        link_type=(LinkType.CALL_CALC, LinkType.CALL_WORK)
    )
    nodes_called = {}
    for link in links_called:
        nodes_called[link.link_label] = [
            link.node.pk,
            link.node.__class__.__name__,
            link.node.node_type,
        ]
    return nodes_called


def get_nodes_caller(pk: int | Node) -> Union[str, Dict[str, Union[List[int], str]]]:
    from aiida.common.links import LinkType

    node = load_node(pk) if isinstance(pk, int) else pk
    links_caller = node.base.links.get_incoming(
        link_type=(LinkType.CALL_CALC, LinkType.CALL_WORK)
    )
    nodes_caller = {}
    for link in links_caller:
        nodes_caller[link.link_label] = [
            link.node.pk,
            link.node.__class__.__name__,
            link.node.node_type,
        ]
    return nodes_caller


def get_node_outputs(pk: int | Node) -> Union[str, Dict[str, Union[List[int], str]]]:
    from aiida.common.links import LinkType

    if pk is None:
        return ""

    node = load_node(pk) if isinstance(pk, int) else pk
    result = ""
    nodes_output = node.base.links.get_outgoing(
        link_type=(LinkType.CREATE, LinkType.RETURN)
    )
    if nodes_output.all():
        result = get_node_recursive(nodes_output.nested())
    else:
        result = {}

    return result


def get_workchain_data(node: Node) -> dict:
    from aiida.common.links import LinkType

    graph_data = {
        "name": node.process_label,
        "uuid": node.uuid,
        "state": node.process_state,
        "nodes": {},
        "links": [],
    }

    links_called = node.base.links.get_outgoing(
        link_type=(LinkType.CALL_CALC, LinkType.CALL_WORK)
    )
    links_input = node.base.links.get_incoming(
        link_type=(LinkType.INPUT_CALC, LinkType.INPUT_WORK)
    )
    nodes_called = []
    for link in links_called:
        graph_data["nodes"][link.node.pk] = {
            "label": f"{link.link_label}-{link.node.pk}",
            "node_type": link.node.node_type,
            "pk": link.node.pk,
            "processPk": link.node.pk,
            "inputs": [],
            "properties": [],
            "outputs": [],
            "position": [0, 0],
            "children": [],
        }
        nodes_called.append(link.node)
    for node in nodes_called:
        links_input = node.base.links.get_incoming(
            link_type=(LinkType.INPUT_CALC, LinkType.INPUT_WORK)
        )
        for link in links_input:
            input_node_links_create = link.node.base.links.get_incoming(
                link_type=(LinkType.CREATE, LinkType.RETURN)
            )
            for link_create in input_node_links_create:
                if link_create.node in nodes_called:
                    graph_data["links"].append(
                        {
                            "from_node": link_create.node.pk,
                            "to_node": node.pk,
                            "from_socket": link_create.link_label,
                            "to_socket": link.link_label,
                        }
                    )
                    input = {"name": link.link_label, "identifier": "any"}
                    output = {"name": link_create.link_label, "identifier": "any"}
                    graph_data["nodes"][node.pk]["inputs"].append(input)
                    graph_data["nodes"][link_create.node.pk]["outputs"].append(output)

    return graph_data


def get_processes_latest(
    pk: int, item_type: str = "called_process"
) -> Dict[str, Dict[str, Union[int, str]]]:
    """Get the latest info of all tasks from the process."""
    import aiida

    tasks = {}
    if pk is None:
        return tasks
    node = aiida.orm.load_node(pk)
    if item_type == "called_process":
        # fetch the process that called by the workgraph
        for link in node.base.links.get_outgoing().all():
            if isinstance(link.node, aiida.orm.ProcessNode):
                tasks[f"{link.link_label}-{link.node.pk}"] = {
                    "pk": link.node.pk,
                    "process_type": link.node.process_type,
                    "state": link.node.process_state.value,
                    "ctime": link.node.ctime,
                    "mtime": link.node.mtime,
                }
    return tasks


def node_to_short_json(workgraph_pk: int, tdata: Dict[str, Any]) -> Dict[str, Any]:
    """Export a node to a rete js node."""

    executor = get_executor_source(tdata)
    tdata_short = {
        "node_type": tdata["metadata"]["node_type"],
        "label": tdata["name"],
        "metadata": [
            ["name", tdata["name"]],
            ["node_type", tdata["metadata"]["node_type"]],
            ["identifier", tdata["identifier"]],
        ],
        "executor": executor,
    }
    process_info = get_processes_latest(workgraph_pk, tdata["name"]).get(
        tdata["name"], {}
    )
    tdata_short["process"] = process_info
    if process_info is not None:
        tdata_short["metadata"].append(["pk", process_info.get("pk")])
        tdata_short["metadata"].append(["state", process_info.get("state")])
        tdata_short["metadata"].append(["ctime", process_info.get("ctime")])
        tdata_short["metadata"].append(["mtime", process_info.get("mtime")])
        tdata_short["inputs"] = get_node_inputs(process_info.get("pk"))
        tdata_short["outputs"] = get_node_outputs(process_info.get("pk"))
    else:
        tdata_short["inputs"] = ""
        tdata_short["outputs"] = ""
    tdata_short["state"] = process_info.get("state", "") if process_info else ""
    return tdata_short


def get_node_summary(node: Node) -> List[List[str]]:
    summary = {
        "table": get_node_summary_table(node),
        "inputs": get_node_inputs(node),
        "outputs": get_node_outputs(node),
        "called": get_nodes_called(node),
        "caller": get_nodes_caller(node),
    }
    return summary


def get_node_summary_table(pk: int | Node) -> List[List[str]]:
    """ """
    from plumpy import ProcessState
    from aiida.orm import ProcessNode

    table = []

    node = load_node(pk) if isinstance(pk, int) else pk

    if isinstance(node, ProcessNode):
        table.append(["type", node.process_label])

        try:
            process_state = ProcessState(node.process_state)
        except (AttributeError, ValueError):
            pass
        else:
            process_state_string = process_state.value.capitalize()

            if process_state == ProcessState.FINISHED and node.exit_message:
                table.append(
                    [
                        "state",
                        f"{process_state_string} [{node.exit_status}] {node.exit_message}",
                    ]
                )
            elif process_state == ProcessState.FINISHED:
                table.append(["state", f"{process_state_string} [{node.exit_status}]"])
            elif process_state == ProcessState.EXCEPTED:
                table.append(["state", f"{process_state_string} <{node.exception}>"])
            else:
                table.append(["state", process_state_string])

    else:
        table.append(["type", node.__class__.__name__])
    table.append(["pk", str(node.pk)])
    table.append(["uuid", str(node.uuid)])
    table.append(["label", node.label])
    table.append(["description", node.description])
    table.append(["ctime", node.ctime])
    table.append(["mtime", node.mtime])

    try:
        computer = node.computer
    except AttributeError:
        pass
    else:
        if computer is not None:
            table.append(["computer", f"[{node.computer.pk}] {node.computer.label}"])

    return table


def time_ago(past_time: datetime) -> str:
    # Get the current time
    now = datetime.now(tzlocal())

    # Calculate the time difference
    delta = relativedelta.relativedelta(now, past_time)

    # Format the time difference
    if delta.years > 0:
        return f"{delta.years}Y ago"
    elif delta.months > 0:
        return f"{delta.months}M ago"
    elif delta.days > 0:
        return f"{delta.days}D ago"
    elif delta.hours > 0:
        return f"{delta.hours}h ago"
    elif delta.minutes > 0:
        return f"{delta.minutes}min ago"
    else:
        return "Just now"


def translate_datagrid_filter_json(raw: str, project) -> dict:
    """
    Convert MUI DataGrid filterModel JSON into AiiDA QueryBuilder filters.
    Supports column filters & quick filter.
    """
    import json

    fm = json.loads(raw)
    filters: dict[str, Any] = {}

    #   DataGrid → QB column
    field_map = {
        "pk": "id",
        "ctime": "ctime",
        "node_type": "node_type",
        "process_label": "attributes.process_label",
        "process_state": "attributes.process_state",
        "exit_status": "attributes.exit_status",
        "exit_message": "attributes.exit_message",
        "paused": "attributes.paused",
        "label": "label",
        "description": "description",
    }

    for item in fm.get("items", []):
        field = item.get("field")
        value = item.get("value")
        operator = item.get("operator", "contains")
        if not value or field not in field_map:
            continue
        col = field_map[field]

        # numeric
        if col == "id":
            try:
                filters[col] = int(value)
            except ValueError:
                continue
        else:
            if operator in ("contains", "equals", "is"):
                filters[col] = {"like": f"%{value}%"}

    # quick filter (space‑separated)
    qf_values = fm.get("quickFilterValues", [])

    if qf_values:
        blocks = []
        for val in qf_values:
            like = {"like": f"%{val}%"}
            block = [{key: like} for key in project]
            block.append({"id": int(val)} if val.isdigit() else {})
            blocks.append({"or": block})
        filters = {"and": [filters, *blocks]} if filters else {"and": blocks}
    return filters


def get_parent_processes(pk: int) -> List[Dict[str, Union[str, int]]]:
    """Get the list of parent processes.
    Use aiida incoming links to find the parent processes.
    the parent process is the process that has a link (type CALL_WORK) to the current process.
    """
    from aiida import orm
    from aiida.common.links import LinkType

    node = orm.load_node(pk)
    parent_processes = [
        {"label": node.process_label, "pk": node.pk, "node_type": node.node_type}
    ]
    links = node.base.links.get_incoming(link_type=LinkType.CALL_WORK).all()
    if len(links) > 0:
        parent_processes.extend(get_parent_processes(links[0].node.pk))
    return parent_processes



from __future__ import annotations
from fastapi import HTTPException
from aiida import orm
import traceback
from aiida_gui.app.node_table import (
    make_node_router,
    process_project,
    projected_data_to_dict_process,
)
from aiida.orm import WorkChainNode
from .utils import get_parent_processes


router = make_node_router(
    node_cls=WorkChainNode,
    prefix="workchain",
    project=process_project,
    get_data_func=projected_data_to_dict_process,
)


@router.get("/api/workchain/{id}")
async def read_workchain(id: int):
    from .utils import get_node_summary, get_workchain_data

    try:

        node = orm.load_node(id)

        content = get_workchain_data(node)
        if content is None:
            print("No workchain data found in the node.")
            return
        summary = get_node_summary(node)
        parent_workflows = get_parent_processes(id)
        parent_workflows.reverse()
        content["summary"] = summary
        content["parent_workflows"] = parent_workflows
        content["processes_info"] = {}
        return content
    except KeyError as e:
        error_traceback = traceback.format_exc()  # Capture the full traceback
        print(error_traceback)
        raise HTTPException(status_code=404, detail=f"Workchain {id} not found, {e}")


@router.get("/api/workchain-state/{id}")
async def read_tasks_state(id: int, item_type: str = "called_process"):
    from aiida_gui.app.utils import get_processes_latest

    try:
        processes_info = get_processes_latest(id, item_type=item_type)
        return processes_info
    except KeyError as e:
        error_traceback = traceback.format_exc()  # Capture the full traceback
        print(error_traceback)
        raise HTTPException(status_code=404, detail=f"Workchain {id} not found, {e}")



from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from aiida.manage import manager
from aiida_gui.app.workchain import router as workchain_router
from aiida_gui.app.task import router as task_router
from aiida_gui.app.process_node import router as process_router
from aiida_gui.app.daemon import router as daemon_router
from aiida_gui.app.data_node import router as datanode_router
from aiida_gui.app.group_node import router as groupnode_router
from fastapi.staticfiles import StaticFiles
from pathlib import Path
import os

from fastapi.responses import FileResponse
from fastapi.exception_handlers import http_exception_handler
from starlette.exceptions import HTTPException as StarletteHTTPException

from pydantic_settings import BaseSettings
from aiida_gui.app.plugin import get_plugins, mount_plugins


class BackendSettings(BaseSettings):
    """
    Settings can be set by setting the environment variables in upper case.
    For example for setting `aiida_gui_profile` one has to export
    the evironment variable `AIIDA_GUI_PROFILE`.
    """

    aiida_gui_profile: str = ""  # if empty aiida uses default profile


backend_settings = BackendSettings()

app = FastAPI()
manager.get_manager().load_profile(backend_settings.aiida_gui_profile)

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allows all origins
    allow_credentials=True,
    allow_methods=["*"],  # Allows all methods
    allow_headers=["*"],  # Allows all headers
)


@app.get("/api", tags=["root"])
async def read_root() -> dict:
    return {"message": "Welcome to AiiDA."}


@app.get("/plugins")
async def list_plugins():
    plugins = get_plugins()
    print(f"Found plugins: {plugins.keys()}")
    plugin_names = [plugin_name for plugin_name in plugins.keys()]
    return {"plugins": plugin_names}


app.include_router(workchain_router)
app.include_router(task_router)
app.include_router(process_router)
app.include_router(datanode_router)
app.include_router(groupnode_router)
app.include_router(daemon_router)
mount_plugins(app)


@app.get("/debug")
async def debug() -> dict:
    return {"loaded_aiida_profile": manager.get_manager().get_profile()}


@app.get("/backend-setting")
async def backend_settings():
    return backend_settings


# Integrating React build into a FastAPI application and serving the build (HTML, CSS, JavaScript) as static files
"""
When navigate to http://127.0.0.1:8000/settings from http://127.0.0.1:8000/ using client-side
routing (i.e., links within your React app), the React Router handles the route /settings
without reloading the page from the server. This is why it works.
However, when you refresh the page at http://127.0.0.1:8000/settings, the browser makes
a request to the FastAPI server for /settings. Since this route isn't defined in FastAPI
(it's a client-side route), the server returns a 404 Not Found error.
so we use the index.html serve all routes except API specific ones, then load all static assets.
"""
backend_dir = Path(__file__).parent
build_dir = backend_dir / "../static"
build_dir = os.getenv("REACT_BUILD_DIR", build_dir)


@app.exception_handler(StarletteHTTPException)
async def _spa_server(req: Request, exc: StarletteHTTPException):
    if exc.status_code == 404:
        return FileResponse(f"{build_dir}/index.html", media_type="text/html")
    else:
        return await http_exception_handler(req, exc)


if os.path.isdir(build_dir):
    app.mount(
        "/static/",
        StaticFiles(directory=build_dir / "static"),
        name="React app static files",
    )

    @app.get("/react-shim.js", include_in_schema=False)
    async def react_shim():
        path = build_dir / "react-shim.js"
        if not path.is_file():
            raise StarletteHTTPException(status_code=404, detail="shim missing")
        return FileResponse(path, media_type="application/javascript")

    @app.get("/react-jsx-runtime-shim.js", include_in_schema=False)
    async def react_jsx_runtime_shim():
        return FileResponse(
            build_dir / "react-jsx-runtime-shim.js", media_type="application/javascript"
        )

    @app.get("/react-router-dom-shim.js", include_in_schema=False)
    async def react_router_dom_shim():
        path = build_dir / "react-router-dom-shim.js"
        if not path.is_file():
            raise StarletteHTTPException(status_code=404, detail="shim missing")
        return FileResponse(path, media_type="application/javascript")

    @app.get("/use-sync-external-store-shim.js", include_in_schema=False)
    async def use_sync_external_store_shim():
        path = build_dir / "use-sync-external-store-shim.js"
        if not path.is_file():
            raise StarletteHTTPException(status_code=404, detail="shim missing")
        return FileResponse(path, media_type="application/javascript")



from __future__ import annotations
from typing import Dict, Any
from fastapi import HTTPException
from aiida_gui.app.node_table import make_node_router
from weas_widget.utils import ASEAdapter
from aiida import orm

project = ["id", "uuid", "ctime", "node_type", "label", "description"]

router = make_node_router(node_cls=orm.Data, prefix="datanode", project=project)


@router.get("/api/datanode/{id}")
async def read_data_node_item(id: int) -> Dict[str, Any]:

    try:
        node = orm.load_node(id)
        content = node.backend_entity.attributes
        content["node_type"] = node.node_type
        # TrajectoryData
        if isinstance(node, orm.TrajectoryData):
            weas_atoms = []
            for i in range(node.numsteps):
                atoms = node.get_step_structure(i).get_ase()
                weas_atoms.append(ASEAdapter.to_weas(atoms))
            content["extras"] = weas_atoms
        return content
    except KeyError:
        raise HTTPException(status_code=404, detail=f"Data node {id} not found")



from __future__ import annotations
from fastapi import APIRouter, Query, Body, HTTPException
from aiida import orm
from typing import Type, Dict, List, Union, Optional


process_project = [
    "id",
    "uuid",
    "ctime",
    "node_type",
    "attributes.process_label",
    "attributes.process_state",
    "attributes.process_status",
    "attributes.exit_status",
    "attributes.exit_message",
    "attributes.paused",
    "label",
    "description",
]


def projected_data_to_dict_process(qb, project):
    """
    Convert the projected data from a QueryBuilder to a list of dictionaries.
    """
    from aiida_gui.app.utils import time_ago

    # Iterate over the results and convert each row to a dictionary
    results = []
    for row in qb.all():
        item = dict(zip(project or [], row))
        # Add computed/presentational fields
        item["pk"] = item.pop("id")
        item["ctime"] = time_ago(item.pop("ctime"))
        item["process_label"] = item.pop("attributes.process_label")
        process_state = item.pop("attributes.process_state")
        item["process_state"] = process_state.title() if process_state else None
        item["process_status"] = item.pop("attributes.process_status")
        item["exit_status"] = item.pop("attributes.exit_status")
        item["exit_message"] = item.pop("attributes.exit_message")
        item["paused"] = item.pop("attributes.paused")
        results.append(item)
    return results


def projected_data_to_dict(qb, project):
    """
    Convert the projected data from a QueryBuilder to a list of dictionaries.
    """
    from aiida_gui.app.utils import time_ago

    # Iterate over the results and convert each row to a dictionary
    results = []
    for row in qb.all():
        item = dict(zip(project or [], row))
        # Add computed/presentational fields
        item["pk"] = item.pop("id")
        item["ctime"] = time_ago(item.get("ctime"))
        results.append(item)
    return results


def make_node_router(
    *,  # force kwargs
    node_cls: Type[orm.Node],  # ⬛  WHICH NODE TYPE
    prefix: str,  # ⬛  URL prefix, e.g. "workgraph" → /api/workgraph-data
    project: Optional[List[str]] = None,
    get_data_func: callable = projected_data_to_dict,
    inclue_delete_route: bool = True,
) -> APIRouter:
    """
    Return an APIRouter exposing GET /…-data, PUT /…-data/{id},
    POST pause/play and DELETE with dry‑run for any AiiDA node subclass.
    """
    from aiida.orm import QueryBuilder
    from aiida.engine.processes.control import (
        pause_processes,
        play_processes,
        kill_processes,
    )
    from aiida.tools import delete_nodes
    from aiida_gui.app.utils import (
        translate_datagrid_filter_json,
    )

    router = APIRouter()

    # -------------------- GET /…-data --------------------
    @router.get(f"/api/{prefix}-data")
    async def read_node_data(
        skip: int = Query(0, ge=0),
        limit: int = Query(15, gt=0, le=500),
        sortField: str = Query(
            "pk", pattern="^(pk|ctime|process_label|state|label|description)$"
        ),
        sortOrder: str = Query("desc", pattern="^(asc|desc)$"),
        filterModel: Optional[str] = Query(None),
    ):
        qb = QueryBuilder()
        filters = (
            translate_datagrid_filter_json(filterModel, project=project)
            if filterModel
            else {}
        )
        qb.append(
            node_cls,
            filters=filters,
            project=project or ["id", "uuid", "ctime", "label", "description"],
            tag="data",
        )

        qb.order_by({"data": {sortField: sortOrder}})
        total = qb.count()
        qb.offset(skip).limit(limit)

        results = get_data_func(qb, project)
        return {"total": total, "data": results}

    # -------------------- PUT /…-data/{id} --------------------
    @router.put(f"/api/{prefix}-data" + "/{id}")
    async def update_node(
        id: int,
        payload: Dict[str, str] = Body(...),
    ):
        try:
            node = orm.load_node(id)
        except Exception:
            raise HTTPException(
                status_code=404, detail=f"{node_cls.__name__} {id} not found"
            )
        allowed = {"label", "description"}
        touched = False
        for k, v in payload.items():
            if k in allowed:
                setattr(node, k, v)
                touched = True
        if not touched:
            raise HTTPException(status_code=400, detail="No updatable fields provided")
        return {"updated": True, "pk": id, **{k: getattr(node, k) for k in allowed}}

    # -------------------- pause / play / delete -------------
    @router.post(f"/api/{prefix}/pause" + "/{id}")
    async def pause(id: int):
        try:
            pause_processes([orm.load_node(id)])
            return {"message": f"Paused {node_cls.__name__} {id}"}
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @router.post(f"/api/{prefix}/play" + "/{id}")
    async def play(id: int):
        try:
            play_processes([orm.load_node(id)])
            return {"message": f"Resumed {node_cls.__name__} {id}"}
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    @router.post(f"/api/{prefix}/kill" + "/{id}")
    async def kill(id: int):
        try:
            kill_processes([orm.load_node(id)])
            return {"message": f"Resumed {node_cls.__name__} {id}"}
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    if inclue_delete_route:

        @router.delete(f"/api/{prefix}/delete" + "/{id}")
        async def delete(
            id: int, dry_run: bool = False
        ) -> Dict[str, Union[bool, str, List[int]]]:
            try:
                deleted, ok = delete_nodes([id], dry_run=dry_run)
                return {
                    "deleted": ok,
                    "message": (
                        f"{'Deleted' if ok else 'Did not delete'} {node_cls.__name__} {id}"
                        + (" [dry‑run]" if dry_run else "")
                    ),
                    "deleted_nodes": list(deleted),
                }
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

    return router



from aiida_gui.app.node_table import (
    make_node_router,
    process_project,
    projected_data_to_dict_process,
)
import traceback
from fastapi import HTTPException
from aiida import orm
from .utils import get_node_summary

router = make_node_router(
    node_cls=orm.ProcessNode,
    prefix="process",
    project=process_project,
    get_data_func=projected_data_to_dict_process,
)


@router.get("/api/process/{id}")
async def read_process(id: int):
    try:
        node = orm.load_node(id)
    except Exception:
        raise HTTPException(status_code=404, detail=f"Process {id} not found")

    data = get_node_summary(node)
    return data


@router.get("/api/process-logs/{id}")
async def read_workgraph_logs(id: int):
    from aiida.cmdline.utils.common import get_workchain_report

    try:
        node = orm.load_node(id)
        report = get_workchain_report(node, "REPORT")
        logs = report.splitlines()
        return logs
    except KeyError as e:
        error_traceback = traceback.format_exc()  # Capture the full traceback
        print(error_traceback)
        raise HTTPException(status_code=404, detail=f"Workgraph {id} not found, {e}")



import click
import subprocess
import os
import signal
from pathlib import Path


def get_package_root():
    """Returns the root directory of the package."""
    current_file = Path(__file__)
    return current_file.parent


def get_pid_file_path():
    """Get the path to the PID file in the desired directory."""
    home_dir = Path.home()
    aiida_daemon_dir = home_dir / ".aiida" / "daemon"

    # Create the directory if it does not exist
    aiida_daemon_dir.mkdir(parents=True, exist_ok=True)

    return aiida_daemon_dir / "web_processes.pid"


@click.group()
def cli():
    """aiida-gui: manage the FastAPI‐based GUI (start/stop)."""
    pass


@cli.command()
@click.option(
    "--watch",
    is_flag=True,
    default=False,
    help="Enable auto-reloading when files change (useful for development).",
)
@click.option(
    "--background",
    "-b",
    is_flag=True,
    default=False,
    help="Run the web application in the background and detach from terminal.",
)
def start(watch, background):
    """Start the web application (FastAPI backend)."""
    pid_file_path = get_pid_file_path()
    command = [
        "uvicorn",
        "aiida_gui.app.api:app",
        "--host",
        "0.0.0.0",
        "--port",
        "8000",
    ]

    if watch:
        command.append("--reload")
        click.echo("Watch mode enabled: The application will reload on file changes.")
    else:
        click.echo(
            "Watch mode disabled: The application will not reload on file changes."
        )

    if background:
        click.echo("Starting the web application in background...")
        # Launch uvicorn in the background
        backend_process = subprocess.Popen(command)
        # Write the PID into our file for later stop
        with open(pid_file_path, "w") as pid_file:
            pid_file.write(f"backend:{backend_process.pid}\n")
        click.echo(f"Web backend started (PID {backend_process.pid}).")
    else:
        click.echo("Starting the web application in foreground. Press Ctrl+C to stop.")
        try:
            # Block until the process is terminated
            subprocess.call(command)
        except KeyboardInterrupt:
            click.echo("\nWeb backend terminated by user.")


@cli.command()
def stop():
    """Stop the web application (terminates any PIDs in the .aiida/daemon file)."""
    pid_file_path = get_pid_file_path()

    if not pid_file_path.exists():
        click.echo("No running web application found.")
        return

    with open(pid_file_path, "r") as pid_file:
        for line in pid_file:
            proc_name, pid = line.strip().split(":")
            try:
                os.kill(int(pid), signal.SIGTERM)
                click.echo(f"Stopped {proc_name} (PID: {pid})")
            except ProcessLookupError:
                click.echo(f"{proc_name} (PID: {pid}) was not found")

    os.remove(pid_file_path)
    click.echo("Cleaned up PID file.")


if __name__ == "__main__":
    cli()
